# -*- coding: utf-8 -*-
"""Settings."""

# File Extensions
CSV_EXTENSION = ".csv"
JSON_EXTENSION = ".json"
PDF_EXTENSION = ".pdf"
TXT_EXTENSION = ".txt"
XLSX_EXTENSION = ".xlsx"
ZIP_EXTENSION = ".zip"

# Defaults
DEFAULT_CSV_DELIMITER = ","
DEFAULT_CSV_DIALECT = "excel"
DEFAULT_CSV_NEWLINE = ""
DEFAULT_CSV_QUOTECHAR = '"'
DEFAULT_FILE_ENCODING = "utf-8"
DEFAULT_TXT_NEWLINE = ""
